
#if LIBRTL2832_PRESENT == 1
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef _WIN32
#include <unistd.h>
#else
#include <Windows.h>
#endif

#include <rtl-sdr.h>
#include <libusb-1.0/libusb.h>
// All libusb callback functions should be marked with the LIBUSB_CALL macro
// to ensure that they are compiled with the same calling convention as libusb.
//
// If the macro isn't available in older libusb versions, we simply define it.

#ifndef LIBUSB_CALL
#define LIBUSB_CALL
#endif

#define BULK_TIMEOUT    0



#include "globdef.h"
#include "uidef.h"
#include "sdrdef.h"
#include "conf.h"
#include "fft1def.h"
#include "screendef.h"
#include "rusage.h"
#include "thrdef.h"
#include "options.h"
#include "vernr.h"

P_RTL2832 rtl2832;
#define MAX_RTL2832_SAMP_RATE  3200000
#define MIN_RTL2832_SAMP_RATE  900001


#define DEFAULT_SAMPLE_RATE		2048000
#define DEFAULT_ASYNC_BUF_NUMBER	32
#define DEFAULT_BUF_LENGTH		(16 * 16384)
#define MINIMAL_BUF_LENGTH		512
#define MAXIMAL_BUF_LENGTH		(256 * 16384)

typedef struct tuner_info {
	/* tuner interface */
	int (*init)(void *);
	int (*exit)(void *);
	int (*set_freq)(void *, uint32_t freq /* Hz */);
	int (*set_bw)(void *, int bw /* Hz */);
	int (*set_gain)(void *, int gain /* dB */);
	int (*set_gain_mode)(void *, int manual);
} rtlsdr_tuner_t;

enum rtlsdr_async_status {
	RTLSDR_INACTIVE = 0,
	RTLSDR_CANCELING,
	RTLSDR_RUNNING
};

struct e4k_pll_params {
	uint32_t fosc;
	uint32_t intended_flo;
	uint32_t flo;
	uint16_t x;
	uint8_t z;
	uint8_t r;
	uint8_t r_idx;
	uint8_t threephase;
};

enum e4k_band {
	E4K_BAND_VHF2	= 0,
	E4K_BAND_VHF3	= 1,
	E4K_BAND_UHF	= 2,
	E4K_BAND_L	= 3,
};


struct e4k_state {
	void *i2c_dev;
	uint8_t i2c_addr;
	enum e4k_band band;
	struct e4k_pll_params vco;
	void *rtl_dev;
};


struct rtlsdr_dev {
	libusb_context *ctx;
	struct libusb_device_handle *devh;
	uint32_t xfer_buf_num;
	uint32_t xfer_buf_len;
	struct libusb_transfer **xfer;
	unsigned char **xfer_buf;
	rtlsdr_read_async_cb_t cb;
	void *cb_ctx;
	enum rtlsdr_async_status async_status;
// rtl demod context
	uint32_t rate; // Hz 
	uint32_t rtl_xtal; // Hz 
// tuner context 
	rtlsdr_tuner_t *tuner;
	uint32_t tun_xtal; //
	uint32_t freq; // Hz 
	int corr; // ppm 
	int gain; // dB
	struct e4k_state e4k_s;
};


static rtlsdr_dev_t *dev_rtlsdr;

char *rtl2832_parm_text[MAX_RTL2832_PARM]={"Sampling speed",
                                     "Serno1",
                                     "Serno2",
                                     "Serno3",
                                     "Check"};

char *rtl2832_parfil_name="par_rtl2832";
void open_rtl2832(void);
void close_rtl2832(void);

static void rtlsdr_callback(unsigned char *buf, uint32_t len, void *ctx)
{
(void) ctx;
unsigned int i;
short int *iz;
iz=(short int*)&timf1_char[timf1p_sdr];
for(i=0; i<len; i++)
  {
  iz[i]=(buf[i]+127)<<8;
  }
timf1p_sdr=(timf1p_sdr+snd[RXAD].block_bytes)&timf1_bytemask;
}

void set_rtl2832_att(void)
{
int i;
i=rtlsdr_set_tuner_gain(dev_rtlsdr, 10*fg.gain);
if (i < 0)
  fprintf(stderr, "WARNING: Failed to set tuner gain  %d %d.\n",i,fg.gain);
      else
       fprintf(stderr, "Tuner gain set to %d dB.\n", i/10);
                          
}

void set_rtl2832_frequency(void)
{
int i;
uint32_t frequency;
frequency=1000000*fg.passband_center;
i=rtlsdr_set_center_freq(dev_rtlsdr, frequency);
if (i < 0)fprintf(stderr, "WARNING: Failed to set center freq.\n");
	else
		fprintf(stderr, "Tuned to %u Hz.\n", frequency);
}



void rtl2832_input(void)
{
#if RUSAGE_OLD == TRUE
int local_workload_counter;
#endif
struct timeval tv1 = { 1, 0 };
int i, j, errcod;
int rxin_local_workload_reset;
unsigned int ii;
char s[80];
double dt1, read_start_time, total_reads;
int timing_loop_counter,timing_loop_counter_max,initial_skip_flag;
int local_att_counter;
int local_nco_counter;
fprintf(stderr,"\nEnter thread rtl input");

#if OSNUM == OSNUM_LINUX
clear_thread_times(THREAD_RTL2832_INPUT);
#endif
#if RUSAGE_OLD == TRUE
local_workload_counter=workload_counter;
#endif
local_att_counter=sdr_att_counter;
local_nco_counter=sdr_nco_counter;
errcod=0;
dt1=current_time();
i=read_sdrpar(rtl2832_parfil_name, MAX_RTL2832_PARM, 
                                     rtl2832_parm_text, (int*)((void*)&rtl2832));
if(i != 0 || rtl2832.check != RTL2832PAR_VERNR)
  {
  errcod=1080;
  goto rtl2832_init_error_exit;
  }
timf1p_sdr=timf1p_pa;  
j=0;
while(sdr == -1)
  {
fprintf(stderr,"\nCall open rtl input");
  
  open_rtl2832();
  if(sdr == -1)
    {
    sprintf(s,"Waiting %.2f", current_time()-dt1);
    lir_text(3,5,s);
    lir_refresh_screen();
    if(kill_all_flag)goto rtl2832_init_error_exit;
    lir_sleep(100000);
    if(j==0)
      {
      clear_screen();
      j=1;
      }
    }
  }

/*
adjusted_sdr_clock=RTL2832_SAMPLING_CLOCK+rtl2832.clock_adjust;
t1=ui.rx_ad_speed-adjusted_sdr_clock/rate_factor[rtl2832.rate_no];
if(fabs(t1/ui.rx_ad_speed) > 0.0001)
  {
  errcod=1091;
  goto rtl2832_error_exit;
  }
ui.rx_ad_speed=adjusted_sdr_clock/rate_factor[rtl2832.rate_no];
*/
set_rtl2832_att();
set_rtl2832_frequency();
fft1_block_timing();
fft1_hz_per_point=(float)ui.rx_ad_speed/fft1_size;
rxin_local_workload_reset=workload_reset_flag;
timing_loop_counter_max=snd[RXAD].interrupt_rate;
timing_loop_counter=2.5*timing_loop_counter_max;
screen_loop_counter_max=0.1*snd[RXAD].interrupt_rate;
if(screen_loop_counter_max==0)screen_loop_counter_max=1;
screen_loop_counter=screen_loop_counter_max;
read_start_time=current_time();
total_reads=0;
initial_skip_flag=1;
if(thread_command_flag[THREAD_SCREEN]!=THRFLAG_NOT_ACTIVE)
  {
  while(thread_status_flag[THREAD_SCREEN]!=THRFLAG_ACTIVE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_IDLE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_SEM_WAIT)
    {
    if(thread_command_flag[THREAD_RTL2832_INPUT] ==
                                           THRFLAG_KILL)goto rtl2832_error_exit;
    lir_sleep(10000);
    }
  }
thread_status_flag[THREAD_RTL2832_INPUT]=THRFLAG_ACTIVE;
while(!kill_all_flag && 
            thread_command_flag[THREAD_RTL2832_INPUT] == THRFLAG_ACTIVE)
  {
  if(rxin_local_workload_reset != workload_reset_flag)
    {
    if(timing_loop_counter <= timing_loop_counter_max)
      {
      rxin_local_workload_reset=workload_reset_flag;
      timing_loop_counter=5*timing_loop_counter_max;
      initial_skip_flag=1;
      }
    }
#if RUSAGE_OLD == TRUE
  if(local_workload_counter != workload_counter)
    {
    local_workload_counter=workload_counter;
    make_thread_times(THREAD_RTL2832_INPUT);
    }
#endif
  if(local_att_counter != sdr_att_counter)
    {
    local_att_counter=sdr_att_counter;
    set_rtl2832_att();
    }
  if(local_nco_counter != sdr_nco_counter)
    {
    local_nco_counter=sdr_nco_counter;
    set_rtl2832_frequency();
    }
  timing_loop_counter--;
  if(timing_loop_counter == 0)
    {
    timing_loop_counter=timing_loop_counter_max;
    if(initial_skip_flag != 0)
      {
      read_start_time=current_time();
      total_reads=0;
      initial_skip_flag=0;
      }
    else
      {
      dt1=current_time()-read_start_time;
      measured_ad_speed=total_reads*snd[RXAD].block_frames/dt1;
      }
    }
  i = libusb_handle_events_timeout(dev_rtlsdr->ctx, &tv1);
  if(i != 0)
    {
    fprintf(stderr, "handle_events returned: %d\n", i);
    }
  if(kill_all_flag)goto rtl2832_error_exit;
  
  while (!kill_all_flag && timf1p_sdr != timf1p_pa)
    {
    total_reads++;
    finish_rx_read();
    }
  }
rtl2832_error_exit:;
dev_rtlsdr->async_status = RTLSDR_CANCELING;
while(dev_rtlsdr->async_status != RTLSDR_INACTIVE) 
  {
  i = libusb_handle_events_timeout(dev_rtlsdr->ctx, &tv1);
  if(i != 0)
    {
    fprintf(stderr, "handle_events returned: %d\n", i);
    }
  if (i < 0) 
    {
    if (i == LIBUSB_ERROR_INTERRUPTED)continue;
    break;
    }
  if(dev_rtlsdr->async_status == RTLSDR_CANCELING ) 
    {
    dev_rtlsdr->async_status=RTLSDR_INACTIVE;
    if (!dev_rtlsdr->xfer)break;
    for(ii = 0; ii < dev_rtlsdr->xfer_buf_num; ++ii) 
      {
      if (!dev_rtlsdr->xfer[ii])continue;
      if (dev_rtlsdr->xfer[ii]->status == LIBUSB_TRANSFER_COMPLETED) 
        {
        libusb_cancel_transfer(dev_rtlsdr->xfer[ii]);
        dev_rtlsdr->async_status = RTLSDR_CANCELING;
        }
      }
    }
  if (RTLSDR_INACTIVE == dev_rtlsdr->async_status)break;
  }
for(i = 0; i < (int)dev_rtlsdr->xfer_buf_num; ++i) 
  {
  if (dev_rtlsdr->xfer[i]) 
    {
    libusb_free_transfer(dev_rtlsdr->xfer[i]);
    }
  }
free(dev_rtlsdr->xfer);
rtlsdr_close(dev_rtlsdr);
if(errcod != 0)lirerr(errcod);
rtl2832_init_error_exit:;
sdr=-1;
thread_status_flag[THREAD_RTL2832_INPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_RTL2832_INPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
fprintf(stderr,"\nLeaving thread rtl input");
} 

                                    


static void LIBUSB_CALL _libusb_callback(struct libusb_transfer *xfer)
{
rtlsdr_dev_t *dev;
dev=(rtlsdr_dev_t *)xfer->user_data;
if (LIBUSB_TRANSFER_COMPLETED == xfer->status)
  {
  if(dev->cb)
    {
    dev->cb(xfer->buffer, xfer->actual_length, dev->cb_ctx);
    }
  libusb_submit_transfer(xfer); /* resubmit transfer */
  } 
}



void open_rtl2832(void)
{
char vendor[256], product[256], serial[256];
int i, k;
int *ise;
unsigned int no_of_rtl;
unsigned int ii;
uint32_t idx;
uint32_t siz1, siz2, siz3;
MEMREF_T ptr;
ise=(int*)(void*)serial;
dev_rtlsdr=NULL;
no_of_rtl = rtlsdr_get_device_count();
for(idx=0; idx<no_of_rtl; idx++)
  {
  rtlsdr_get_device_usb_strings(idx, vendor, product, serial);
  if( rtl2832.sernum1 == ise[0] &&
      rtl2832.sernum2 == ise[1] &&
      rtl2832.sernum3 == ise[2])
    {  
    sdr=rtlsdr_open(&dev_rtlsdr, idx);
    if(kill_all_flag)return;
    if(sdr >= 0)
      {
      i=rtlsdr_set_sample_rate(dev_rtlsdr, (uint32_t)rtl2832.sampling_speed);
      if(i < 0)
        {
        lirerr(1339);
opnerr:;        
	rtlsdr_close(dev_rtlsdr);
        return;
        }

// Reset endpoint before we start reading from the RTL2832 (mandatory)
      i=rtlsdr_reset_buffer(dev_rtlsdr);
      if(i < 0)
        {
        lirerr(1340);
        goto opnerr;
        }
      k=snd[RXAD].block_bytes/2;
      if(k < 512)k=512;
      dev_rtlsdr->cb=rtlsdr_callback;
      dev_rtlsdr->cb_ctx=NULL;
      dev_rtlsdr->xfer_buf_num=32; // no of buffers
      dev_rtlsdr->xfer_buf_len=k;
      siz1=sizeof(struct libusb_transfer *);
      siz2=sizeof(unsigned char *);
      siz3=dev_rtlsdr->xfer_buf_len;
      dev_rtlsdr->xfer = malloc(dev_rtlsdr->xfer_buf_num*(siz1+siz2+siz3));
      ptr=(MEMREF_T)dev_rtlsdr->xfer;
      for(ii = 0; ii < dev_rtlsdr->xfer_buf_num; ++ii)
        {
        dev_rtlsdr->xfer[ii] = libusb_alloc_transfer(0);
        ptr+=siz1;
        }
      dev_rtlsdr->xfer_buf=(unsigned char **)ptr;
      ptr+=dev_rtlsdr->xfer_buf_num*siz2;
      for(ii = 0; ii < dev_rtlsdr->xfer_buf_num; ++ii)
        {
        dev_rtlsdr->xfer_buf[ii]=(unsigned char *)ptr;
        ptr+=siz3;
        }
      for(ii = 0; ii < dev_rtlsdr->xfer_buf_num; ++ii)
        {
        libusb_fill_bulk_transfer(dev_rtlsdr->xfer[ii],
                                  dev_rtlsdr->devh,
                                  0x81,
                                  dev_rtlsdr->xfer_buf[ii],
                                  dev_rtlsdr->xfer_buf_len,
                                  _libusb_callback,
                                  (void *)dev_rtlsdr,
                                  BULK_TIMEOUT);
        libusb_submit_transfer(dev_rtlsdr->xfer[ii]);
        }
      dev_rtlsdr->async_status = RTLSDR_RUNNING;
      rtlsdr_set_tuner_gain_mode(dev_rtlsdr, 1);
      return;
      }
    }
  }
}

void init_rtl2832(void)
{
uint32_t resamp_ratio;
FILE *rtl2832_file;
char s[120];
char vendor[256], product[256], serial[256];
int *ise, *sdr_pi;;
int i, line, no_of_rtl, devno;
double actual;
ise=(int*)(void*)serial;
line=0;
no_of_rtl = rtlsdr_get_device_count();
if (no_of_rtl == 0) 
  {
  lir_text(5,5,"RTL2832 not found.");
  lir_text(5,7,press_any_key);
  await_keyboard();
  return;
  }
line=2;  
for(i=0; i<no_of_rtl; i++)
  {
  rtlsdr_get_device_usb_strings(i, vendor, product, serial);
  sprintf(s," %2d: %s   MFG:%s, PROD:%s, SN: %s", i, 
                           rtlsdr_get_device_name(i), vendor, product, serial);
  lir_text(3,line,s);
  line++;
  }
line++;
lir_text(3, line, "Select device by line number:");
devno=lir_get_integer(32,line,2,0,no_of_rtl-1);
rtlsdr_get_device_usb_strings(devno, vendor, product, serial);
rtl2832.sernum1=ise[0];
rtl2832.sernum2=ise[1];
rtl2832.sernum3=ise[2];
line+=2;
i=rtlsdr_open(&dev_rtlsdr, (uint32_t)devno);
if(i<0)lirerr(231176);
getspeed:;
sprintf(s,"Set sampling speed in Hz (%d to %d) =>",
                                    MIN_RTL2832_SAMP_RATE,MAX_RTL2832_SAMP_RATE);
lir_text(3,line,s);

rtl2832.sampling_speed=lir_get_integer(4+strlen(s),line,7,
                                MIN_RTL2832_SAMP_RATE,MAX_RTL2832_SAMP_RATE);
resamp_ratio=(dev_rtlsdr->rtl_xtal * (double)4194304)/rtl2832.sampling_speed;
resamp_ratio &= ~3;
actual = (dev_rtlsdr->rtl_xtal * (double)4194304) / resamp_ratio;
sprintf(s,"Actual=%f. Use this sampling rate? (Y/N)",actual);
lir_text(3,line+1,s);
to_upper_await_keyboard();
if(kill_all_flag)
  {
  rtlsdr_close(dev_rtlsdr);
  return;
  }
clear_lines(line,line+1);
if(lir_inkey != 'Y')goto getspeed;
i=rtlsdr_set_sample_rate(dev_rtlsdr, (uint32_t)rtl2832.sampling_speed);
rtlsdr_close(dev_rtlsdr);
if(i < 0)
  {
  lir_text(3,line,"Failed to set sampling rate");
  lir_text(3,line+1,press_any_key);
  await_keyboard();
  clear_lines(line,line+1);
  goto getspeed;
  }
rtl2832_file=fopen("par_rtl2832","w");
if(rtl2832_file == NULL)
  {
  lirerr(381268);
  return;
  }
rtl2832.check=RTL2832PAR_VERNR;
sdr_pi=(int*)(&rtl2832);
for(i=0; i<MAX_RTL2832_PARM; i++)
  {
  fprintf(rtl2832_file,"%s [%d]\n",rtl2832_parm_text[i],sdr_pi[i]);
  }
parfile_end(rtl2832_file);
ui.rx_addev_no=RTL2832_DEVICE_CODE;
ui.rx_ad_speed=rtl2832.sampling_speed;
ui.rx_input_mode=IQ_DATA+DIGITAL_IQ;
ui.rx_rf_channels=1;
ui.rx_ad_channels=2;
ui.rx_admode=0;
}

int display_rtl2832_parm_info(int *line)
{
char s[80];
int errcod;
errcod=read_sdrpar(rtl2832_parfil_name, MAX_RTL2832_PARM, 
                               rtl2832_parm_text, (int*)((void*)&rtl2832));
if(errcod == 0)
  {
  settextcolor(7);
  sprintf(s,"sampling rate = %i ", rtl2832.sampling_speed);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
//  line[0]+=1;
  }
return (errcod);
}
#endif
